from skilletlib import Panoply
from skilletlib.exceptions import PanoplyException

from .base import BaseCommand
from ..decorators import require_ngfw_connection_params
from ..decorators import require_panoply_connection


class BaselineCommand(BaseCommand):
    sli_command = "baseline"
    short_desc = "Revert configuration to a near factory default, but retaining admin auth and network configuration"
    no_skillet = True

    @require_ngfw_connection_params
    @require_panoply_connection
    def run(self, pan: Panoply):
        try:
            saved_configs = pan.list_saved_configurations()
            if "skillet_baseline.xml" in saved_configs:
                pan.load_config("skillet_baseline.xml")
            else:
                pan.load_baseline()

            print(f'Successfully reverted {self.sli.context["TARGET_IP"]} to baseline config.')

            if self.sli.commit:
                print("Committing configuration...")
                pan.commit()
                print("Success...")

        except PanoplyException as pe:
            print(pe)
            exit(1)



from skilletlib.exceptions import PanoplyException

from sli.decorators import require_ngfw_connection_params
from sli.decorators import require_panoply_connection
from .base import BaseCommand
from sli.errors import InvalidArgumentsException
from sli.errors import SLIException
from sli.tools import load_config_file, format_xml_string, get_input, load_app_skillet


class DiffCommand(BaseCommand):
    sli_command = "diff"
    short_desc = "Get the differences between two config versions: candidate, running, previous running, etc"
    no_skillet = True
    capture_var = None
    pan = None
    help_text = """
        Diff module requires 0 to 3 arguments.

        - 0 arguments: Diff running config from previous running config
        - 1 argument: Diff previous config or named config against specified running config
        - 2 arguments: Diff first arg named config against second arg named config
        - 3 arguments: Diff first arg named config against second arg named config and save diffs into the context

        A named config can be either a stored config, candidate, running or a number.
        Positive numbers must be used to specify iterations, 1 means 1 config revision ago

        Example: Get diff between running config and previous running config in set cli format

            user$ sli diff 1 -uc -of set

        Example: Get diff between running config and the candidate config in xml format

            user$ sli diff running candidate -uc -of xml

        Example: Get diff between running config and the candidate config in skillet format

            user$ sli diff running candidate -uc -of skillet

        Example: Get the diff of all changes made to this device using the autogenerated 'baseline' config, which is
        essentially blank.

            user$ sli diff baseline -uc -of xml

        Example: Get the diff between the second and third most recent running configs

            user$ sli diff 3 2

        Example: Get a diff and save as 'candidate_diff' into the context

            user$ sli diff running candidate candidate_diff -uc

        Example: Get a diff from running and a local config file, save as out.xml. Note the 'file:' prefix.

            user$ sli diff running file:test-file.xml candidate_diff -uc -o out.xml

        Example: Get a diff between two local saved configs and same as diff.out. Note the '--offline' flag.

            user$ sli diff running test-file.xml test-file-2.xml --offline -o diff.out
    """

    def _parse_args(self) -> None:
        """
        handle arguments for this Command.
        """

        def fixup(x):
            return f"-{x}" if x.isdigit() else x

        if len(self.args) == 1:
            latest_name = "running"
            source_name = fixup(self.args[0])

        elif len(self.args) == 2:
            source_name = fixup(self.args[0])
            latest_name = fixup(self.args[1])

        elif len(self.args) == 3:
            source_name = fixup(self.args[0])
            latest_name = fixup(self.args[1])
            self.capture_var = self.args[2]

        elif len(self.args) > 3:

            raise InvalidArgumentsException("Too many arguments")

        else:
            # If no arguments supplied, assume diff previous running vs running
            latest_name = "running"
            source_name = "-1"

        self.source_name = source_name
        self.latest_name = latest_name

    def _handle_snippets(self, snippets: list, vars: list) -> None:
        output = ""

        if self.sli.output_format == "xml":
            for obj in snippets:
                element = format_xml_string(obj["element"], indent=2)
                output += f'name: {obj["name"]}\n'
                output += f'xpath: {obj["xpath"]}\n'
                output += f"element: |-\n{element}\n\n"

        elif self.sli.output_format == "set":
            output = "\n".join(snippets)

        else:

            for snippet in snippets:
                snippet["element"] = format_xml_string(snippet["element"], indent=6)

            panos_skeleton = load_app_skillet("panos_skillet_skeleton")

            skillet_name = get_input("Skillet Name:", "my_skillet")
            skillet_label = get_input("Skillet Label:", "my label")
            skillet_description = get_input("Skillet Description:", "my description")
            skillet_output = panos_skeleton.execute(
                {
                    "snippets": snippets,
                    "skillet_name": skillet_name,
                    "skillet_label": skillet_label,
                    "skillet_description": skillet_description,
                }
            )

            if not panos_skeleton.success:
                raise SLIException("Could not generate Skillet output")

            output = skillet_output["template"]

        self._handle_outfile(output)
        print(output)

    def _get_vars(self) -> list:
        return []

    def _get_snippets(self) -> list:
        """
        Internal method to actually perform the diff operation.
        """

        previous_config = load_config_file(self.source_name, self.pan)
        latest_config = load_config_file(self.latest_name, self.pan)
        if self.sli.output_format == "set":
            snippets = self.pan.generate_set_cli_from_configs(previous_config, latest_config)
        else:
            snippets = self.pan.generate_skillet_from_configs(previous_config, latest_config)

        return snippets

    def _handle_outfile(self, output: str) -> None:

        out_file = self.sli.options.get("out_file")
        if output and out_file:
            with open(out_file, "w") as f:
                f.write(output)

    def _update_context(self, diff: list) -> None:

        # Update context if using context
        if self.sli.cm.use_context and self.capture_var is not None:
            self.sli.context[self.capture_var] = diff
            print(f"Output added to context as {self.capture_var}")

    @require_ngfw_connection_params
    @require_panoply_connection
    def run(self, pan):
        """Get a diff of running and candidate configs"""

        self.pan = pan

        try:
            self._parse_args()

        except InvalidArgumentsException:
            self._print_usage()
            return

        try:
            diff = self._get_snippets()
            vars_list = self._get_vars()
            self._update_context(diff)
            self._handle_snippets(diff, vars_list)

        except PanoplyException as pe:
            print(pe)
            exit(1)

        except SLIException as sle:
            print(sle)
            exit(1)



from sli.tools import print_table
from .base import BaseCommand


class ListContext(BaseCommand):
    sli_command = 'list_context'
    short_desc = 'List all available contexts'
    no_skillet = True
    no_context = True

    def run(self):
        contexts = self.sli.cm.get_contexts()
        for i in range(len(contexts)):
            contexts[i]['encrypted'] = 'Yes' if contexts[i]['encrypted'] else ''
        print_table(contexts,
                    {
                        "Name": "name",
                        "Encrypted": "encrypted",
                    })



from sli.decorators import load_variables
from sli.decorators import require_single_skillet
from sli.decorators import require_skillet_type
from .base import BaseCommand

import os

"""
Render a template and save it to a file
"""


class TemplateCommand(BaseCommand):
    sli_command = 'template'
    short_desc = 'Render a template and safe it to a file'
    help_text = """

        Render a template and save it to a file. Templates will be created
        named as they are named in the skillet. If an out_directory is not
        specified, the current working directory will be used.

        Usage:
            sli template -n template_name out_directory
"""

    @require_single_skillet
    @require_skillet_type('template')
    @load_variables
    def run(self):

        if len(self.args) > 1:
            print(self.help_text)
            return
        elif len(self.args) == 1:
            out_directory = self.args[0]
        else:
            out_directory = ""

        output = self.sli.skillet.execute(self.sli.context)
        output = output["template"]

        out_file = self.sli.options.get("out_file", False)
        if out_file:

            out_path = os.path.sep.join([out_directory, out_file]) if len(out_directory) else out_file
            with open(out_path, 'w') as f:
                f.write(output)
            print(f"Template written to {out_path}")
        else:
            print(output)



from .base import BaseCommand
from sli.errors import InvalidArgumentsException
import re
import os
import shutil
from skilletlib import SkilletLoader


class UpdateSkilletVars(BaseCommand):
    sli_command = "update_skillet_vars"
    short_desc = "Finds and updates all template variables in a Skillet"
    no_skillet = True
    no_context = True
    help_text = """

    This command is useful when building or modifying an existing skillet. You can edit the
    XML elements or other templated attributes with Jinja {{ variables }} then run this command
    to update the list of variables with your new additions.

    Usage:
        sli update_skillet_vars yourfile.skillet.yaml
    """

    def run(self):

        if not len(self.args) == 1:
            raise InvalidArgumentsException("Must specify single argument of file to update")
        file_name = self.args[0]
        file_bkp = file_name + ".bkp"
        file_buffer = []  # Store the file as a list of lines in this buffer to replay later

        # Load skillet, extract variables and determine missing
        sl = SkilletLoader()
        skillet = sl.load_skillet(file_name)
        vars_specified = [x['name'] for x in skillet.variables]
        missing_vars = [x for x in skillet.get_declared_variables() if x not in vars_specified]
        if len(missing_vars) < 1:
            print("Did not find any variables in snippets that were not declared")
            return

        # Load the file and find the insertion point of new variables
        with open(file_name, "r") as f:
            i = 1
            variables_start = None
            variables_end = None
            in_var_section = False
            indent_dash = 2
            indent_text = 4
            for line in f.readlines():

                file_buffer.append(line)

                # Locate the beginning of the variables section
                if line.strip() == "variables:":
                    variables_start = i
                    in_var_section = True

                # Locate the end of the variables section
                elif variables_start and line.strip() == "snippets:":
                    variables_end = i
                    in_var_section = False

                # Find first line of existing variable block and infer spacing
                elif in_var_section:
                    if re.search(".*-.*name:", line):
                        indent_dash = line.index("-")
                        indent_text = line.index("n")

                i += 1

        # Backup the file in case of issue
        if os.path.exists(file_bkp):
            os.remove(file_bkp)
        shutil.copy(file_name, file_bkp)

        # Overwrite the user submitted file
        with open(file_name, "w") as f:
            i = 1
            for line in file_buffer:

                # End of variables section, append missing_vars here
                if i == variables_end:
                    for mv in missing_vars:
                        f.write(f"{' ' * indent_dash}- name: {mv}\n")
                        f.write(f"{' ' * indent_text}description: {mv}\n")
                        f.write(f"{' ' * indent_text}default: \n")
                        f.write(f"{' ' * indent_text}type_hint: text\n\n")

                f.write(line)
                i += 1

        # Delete the backup file
        os.remove(file_bkp)



from skilletlib import Panoply
from skilletlib.exceptions import PanoplyException

from .base import BaseCommand
from ..decorators import require_ngfw_connection_params
from ..decorators import require_panoply_connection


class CommitCommand(BaseCommand):
    sli_command = "commit"
    short_desc = "Commit the Candidate configuration"
    no_skillet = True

    @require_ngfw_connection_params
    @require_panoply_connection
    def run(self, pan: Panoply):
        try:
            print("Committing configuration...")
            pan.commit()
            print("Success...")

        except PanoplyException as pe:
            print(pe)
            exit(1)



from .base import BaseCommand
from sli.decorators import require_ngfw_connection_params, require_ngfw_ssh_session
from sli.progressBar import ProgressBar
from sli.tools import get_password_input

import re


class LoadSet(BaseCommand):
    sli_command = "load_set"
    short_desc = "Load set commands in a file against a live NGFW to ensure no errors"
    no_skillet = True
    help_text = """
        Load set commands into an NGFW from a specified file.
        fails out on any set command errors

        Example usage with progress bar and timer:
            sli load_set -uc set_commands.txt

        Example usage printing out commands as they are processed:
            sli load_set -uc set_commands.txt -v
"""

    @require_ngfw_connection_params
    @require_ngfw_ssh_session
    def run(self, ssh):

        # Load script as a list of lines off of disk
        if not len(self.args) == 1:
            print(self.help_text)
            exit()
        script_name = self.args[0]
        with open(script_name, 'r') as f:
            script = [x.strip() for x in f.read().split('\n')]

        # Filter out set non executable lines and verify contents
        commands = [x for x in script if not x.startswith('#') and len(x)]
        invalid_commands = [x for x in commands if not x.startswith("set ") and not x.startswith("delete ")]
        if len(invalid_commands):
            print("Following commands are not set commands: ")
            for command in invalid_commands:
                print(f"  - {command}")

        # Execute line by line until end of script
        print("Starting set command load...")
        total = len(commands)
        ssh.config_mode()
        i = 0
        if not self.sli.verbose:
            pb = ProgressBar(prefix="Applying commands")
        for command in commands:

            # Check if a password will be required as the next input
            if re.match("^set mgt-config users.*password$", command.strip()):
                username = command.strip().split(" ")[3]
                if not self.sli.verbose:
                    pb.pause()
                password = get_password_input(f"Password for user {username}")
                ssh.extended_set_command(command, expect="Enter password")
                ssh.extended_set_command(password, expect="Confirm password")
                ssh.extended_set_command(password)
                continue

            # Process normal command
            if not ssh.set_command(command):
                print('\n' + ssh.get_error_text())
                print("Errors occurred while loading set commands.")
                return
            i += 1
            percent = "{0:.1f}".format(100 * (i / total))
            if self.sli.verbose:
                print(command)
            else:
                pb.update(percent)

        if not self.sli.verbose:
            pb.complete()

        print('Set commands successfully loaded')



from sli.commands.base import BaseCommand
from pathlib import Path
from pkgutil import iter_modules
from inspect import isclass
from importlib import import_module

# Disable SSL warning
import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

# Load all command modules from this package

package_dir = Path(__file__).resolve().parent
for (_, module_name, _) in iter_modules([str(package_dir)]):

    # Load each module individually
    module = import_module(f"{__name__}.{module_name}")
    for attr_name in dir(module):
        attr = getattr(module, attr_name)

        # Evaluate each class found in the module
        if attr is BaseCommand:
            continue
        if isclass(attr):
            if not len(getattr(attr, 'sli_command', '')) > 0:
                continue
            if not issubclass(attr, BaseCommand):
                raise ImportError(f'Command module {attr} must subclass BaseCommand')
            globals()[attr_name] = attr



from sli.decorators import load_variables
from sli.decorators import require_single_skillet
from sli.decorators import require_skillet_type
from sli.decorators import require_config
from .base import BaseCommand
from sli.tools import merge_xml_into_config

from lxml import etree

"""
Generate output of a panos skillet and write to file
"""


class PreviewCommand(BaseCommand):
    sli_command = "preview"
    short_desc = "Generate output of a panos skillet and write to file"
    help_text = """

        Load and run a panos configuration skillet, saving the output to
        disk in a specified directory as opposed to configuring NGFW

        Usage:
            sli preview -n configuration_skillet -o output_file.xml
"""

    @require_single_skillet
    @require_skillet_type("panos", "panorama")
    @load_variables
    @require_config
    def run(self, config):

        out_file = self.sli.options.get("out_file", "out.xml")
        snippets = self.sli.skillet.get_snippets()
        for snippet in snippets:

            if not snippet.should_execute(self.sli.context):
                continue
            meta = snippet.render_metadata(self.sli.context)
            xpath = snippet.metadata.get("xpath")
            element = meta.get("element")
            child_tag = xpath.split("/")[-1]
            child_xml = etree.fromstring(f"<{child_tag}>{element}</{child_tag}>")
            merge_xml_into_config(xpath, config, child_xml)

        config.write(out_file, encoding="utf-8", pretty_print=True, xml_declaration=True)



from .ansible_role_from_diff import AnsibleRoleCommand
from ..decorators import require_single_skillet
from ..decorators import require_skillet_type
from ..errors import InvalidArgumentsException
from ..errors import SLIException
from ..tools import get_input


class SkilletToAnsibleRoleCommand(AnsibleRoleCommand):
    sli_command = "ansible_role_from_skillet"
    short_desc = "Builds an Ansible Role from the a Skillet"
    no_skillet = False
    capture_var = None
    pan = None
    suppress_output = True
    help_text = """
        ansible_role_from_skillet creates a complete Ansible Collection and Role directory structure from a skillet.

        Example: Create a collection in the /tmp/roles directory using the test_skillet from the /tmp/skillets directory

            user$ sli ansible_role_from_skillet --name test_skillet -sd /tmp/skillets -od /tmp/roles
    """

    def _parse_args(self) -> None:
        pass

    def _get_snippets(self) -> list:
        """
        Internal method to actually perform the diff operation.
        """

        skillet_dict = self.sli.skillet.skillet_dict
        return skillet_dict["snippets"]

    def _get_vars(self) -> list:
        return self.sli.skillet.skillet_dict["variables"]

    def _update_context(self, diff: list) -> None:
        pass

    def _get_version(self):
        """
        Sadly, we can't get this directly from the loaded skillet programmatically...
        """
        return get_input("PAN-OS Version to target", "10.0")

    @require_single_skillet
    @require_skillet_type("panos", "panorama")
    def run(self):
        """Get a diff of running and candidate configs"""

        try:
            self._parse_args()

        except InvalidArgumentsException:
            self._print_usage()
            return

        try:
            diff = self._get_snippets()
            variable_list = self._get_vars()
            self._update_context(diff)
            self._handle_snippets(diff, variable_list)

        except SLIException as sle:
            print(sle)
            exit(1)
